/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#include "fggl/math/shapes.hpp"
#include <vector>

namespace fggl::math::phs3d {

	void AABB::emtpy() {
		min = {FLT_MAX, FLT_MAX, FLT_MAX};
		max = {-FLT_MAX, -FLT_MAX, -FLT_MAX};
	}

	void AABB::add(const math::vec3 &point) {
		min = minElm(min, point);
		max = maxElm(max, point);
	}

	AABB AABB::fromPoints(const std::vector<math::vec3> &points) {
		AABB box;
		for (const auto &point : points) {
			box.add(point);
		}
		return box;
	}

	void AABB::set(const AABB &other, const math::mat4 &m) {
		// TODO this needs testing, I'm not sure if our vectors are oriented the same as 9.4.4
		min = max = m[3]; // should be the translation component of the matrix

		// this feels like something that should be vectorizable...
		for (int i = 0; i < 3; i++) {
			for (int j = 0; j < 3; j++) {
				if (m[i][j] > 0.0f) {
					min[j] += m[i][j] * other.min[j];
					max[j] += m[i][j] * other.max[j];
				} else {
					min[j] += m[i][j] * other.max[j];
					max[j] += m[i][j] * other.min[j];
				}
			}
		}
	}

	Plane Plane::fromPoints(const math::vec3 p1, const math::vec3 p2, const math::vec3 p3) {
		const auto e3 = p2 - p1;
		const auto e1 = p3 - p2;
		auto normal = glm::normalize(glm::cross(e3, e1));
		auto d = glm::dot(p2, normal);
		return {normal, d};
	}

	static math::vec3 bestFitNormal(const std::vector<math::vec3> &points) {
		assert(!points.empty());

		math::vec3 result;

		math::vec3 p = points.back();
		for (std::size_t i = 0; i < points.size(); ++i) {
			math::vec3 c = points[i];

			result.x += (p.z + c.z) * (p.y - c.y);
			result.y += (p.x + c.x) * (p.z - c.z);
			result.z += (p.y + c.y) - (p.x - c.x);

			p = c;
		}

		return glm::normalize(result);
	};

	static float bestFitD(const std::vector<math::vec3> &points, glm::vec3 normal) {
		math::vec3 sum;
		for (auto &point : points) {
			sum += point;
		}
		sum *= 1.0F / points.size();
		return glm::dot(sum, normal);
	}

	const char X = 0;
	const char Y = 1;
	const char Z = 2;

	struct bary_axis {
		const char a;
		const char b;
	};

	static bary_axis baryCalcAxis(const math::vec3 &normal) {
		if ((fabs(normal.x) >= fabs(normal.y)) && (fabs(normal.x) >= fabs(normal.z))) {
			// discard x
			return {Y, Z};
		} else if (fabs(normal.y) >= fabs(normal.z)) {
			// discard y
			return {Z, X};
		} else {
			// discard z
			return {X, Y};
		}
	}

	bool Triangle::CartToBarycentric(const math::vec3 &cart, Barycentric &outVal) {
		// everything is const because I'm paying the compiler is smarter than me...

		const auto d1 = v[1] - v[0];
		const auto d2 = v[2] - v[1];
		const auto n = glm::cross(d1, d2);

		// 0 = x, 1 = y, 2 = z
		const auto ax = baryCalcAxis(n);

		// first axis
		const float u1 = v[0][ax.a] - v[2][ax.a];
		const float u2 = v[1][ax.a] - v[2][ax.a];
		const float u3 = cart[ax.a] - v[0][ax.a];
		const float u4 = cart[ax.a] - v[2][ax.a];

		// second axis
		const float v1 = v[0][ax.b] - v[2][ax.b];
		const float v2 = v[1][ax.b] - v[2][ax.b];
		const float v3 = cart[ax.b] - v[0][ax.b];
		const float v4 = cart[ax.b] - v[2][ax.b];

		const float denom = v1 * u2 - v2 * u1;
		if (denom == 0.0f) {
			return false;
		}

		// finally, we can work it out
		const float oneOverDenom = 1.0f / denom;
		outVal.b[0] = (v4 * u2 - v2 * u4) * oneOverDenom;
		outVal.b[1] = (v1 * u3 - v3 * u1) * oneOverDenom;
		outVal.b[2] = 1.0f - outVal.b[0] - outVal.b[1];
		return true;
	}

	bool Triangle::CartToBarycentric2(const math::vec3 &cart, Barycentric &outVal) {
		const auto e1 = v[2] - v[1];
		const auto e2 = v[0] - v[2];
		const auto e3 = v[1] - v[0];

		const auto d1 = cart - v[0];
		const auto d2 = cart - v[1];
		const auto d3 = cart - v[2];

		const auto normal = glm::normalize(glm::cross(e1, e2));
		const auto denom = glm::dot(glm::cross(e1, e2), normal);
		assert(denom != 0.0f);

		outVal.b[0] = glm::dot(glm::cross(e1, d3), normal) / denom;
		outVal.b[1] = glm::dot(glm::cross(e2, d1), normal) / denom;
		outVal.b[2] = glm::dot(glm::cross(e3, d2), normal) / denom;
		return true;
	}

} // namespace fggl::math::phys3d