/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef FGGL_GFX_OGL_RENDERER_HPP
#define FGGL_GFX_OGL_RENDERER_HPP

#include <fggl/data/model.hpp>

#include <fggl/gfx/ogl/backend.hpp>
#include <fggl/gfx/ogl/shader.hpp>
#include "fggl/gfx/ogl4/models.hpp"
#include "fggl/gfx/ogl4/canvas.hpp"
#include "fggl/gfx/ogl4/debug.hpp"

#include <fggl/gfx/paint.hpp>

#include <memory>
#include <vector>

namespace fggl::gfx {

	using GlFunctionLoader = GLADloadproc;

	enum GlRenderType {
		triangles = GL_TRIANGLES,
		triangle_strip = GL_TRIANGLE_STRIP
	};

	struct GlRenderToken {
		constexpr static const char name[] = "RenderToken";
		GLuint vao;
		GLuint buffs[2];
		GLuint idxOffset;
		GLsizei idxSize;
		GLuint pipeline;
		GLuint restartVertex;
		GlRenderType renderType = triangles;
	};

	/**
	 * Class responsible for managing the OpenGL context.
	 *
	 * This class manages the OpenGL resources, all OpenGL resources are bound by its lifetime, and it is ultimately
	 * the class responsible for interacting with OpenGL. It is bound by the window it belongs to, and therefore if
	 * that window disappears so does this object.
	 */
	class OpenGL4Backend : public Graphics {
		public:
			explicit OpenGL4Backend(data::Storage *storage, gui::FontLibrary *fonts, assets::AssetManager *assets, GlFunctionLoader loader);
			~OpenGL4Backend() override = default;

			// copy bad
			OpenGL4Backend(const OpenGL4Backend &) = delete;
			OpenGL4Backend &operator=(const OpenGL4Backend &) = delete;

			// move (probably) bad
			OpenGL4Backend(OpenGL4Backend &&) = delete;
			OpenGL4Backend &&operator=(OpenGL4Backend &&) = delete;

			/**
			 * Clear the backing buffer.
			 */
			void clear() override;

			/**
			 * Notify the graphics system the framebuffer size has changed.
			 *
			 * @param width the new framebuffer width
			 * @param height the new framebuffer height
			 */
			void resize(int width, int height) override;

			/**
			 * Perform a 2D rendering pass.
			 *
			 * @param paint the objects to paint on screen
			 */
			void draw2D(const Paint &paint) override;

			/**
			 * Perform a 3D rendering pass.
			 *
			 * @param world the world to render
			 */
			void drawScene(entity::EntityManager &world) override;

			/**
			 * Get the 2D canvas bounds.
			 *
			 * @return the canvas bounds
			 */
			inline Bounds canvasBounds() override {
				return m_canvasRenderer->bounds();
			}

		private:
			std::unique_ptr<ogl4::StaticModelRenderer> m_modelRenderer;
			std::unique_ptr<ogl4::CanvasRenderer> m_canvasRenderer;
			std::unique_ptr<ogl4::DebugRenderer> m_debugRenderer;
			std::unique_ptr<ShaderCache> m_cache;
			std::shared_ptr<ogl::Shader> m_canvasPipeline;
			data::Storage *m_storage;
			gui::FontLibrary *m_fontLibrary;
	};

};  // namespace fggl::gfx

#endif
